@@ -1,23 +1,37 @@
 """All pytest-django fixtures"""
-from typing import Any, Callable, Generator, Iterable, List, Optional, Tuple, Union
+from __future__ import annotations
+
 import os
 from contextlib import contextmanager
 from functools import partial
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    ClassVar,
+    Generator,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    Tuple,
+    Union,
+)
 
 import pytest
 
 from . import live_server_helper
 from .django_compat import is_django_unittest
-from .lazy_django import skip_if_no_django, get_django_version
+from .lazy_django import skip_if_no_django
+
 
-TYPE_CHECKING = False
 if TYPE_CHECKING:
-    from typing import Literal
-
     import django
 
-    _DjangoDbDatabases = Optional[Union["Literal['__all__']", Iterable[str]]]
-    _DjangoDb = Tuple[bool, bool, _DjangoDbDatabases]
+
+_DjangoDbDatabases = Optional[Union[Literal["__all__"], Iterable[str]]]
+_DjangoDbAvailableApps = Optional[List[str]]
+# transaction, reset_sequences, databases, serialized_rollback, available_apps
+_DjangoDb = Tuple[bool, bool, _DjangoDbDatabases, bool, _DjangoDbAvailableApps]
 
 
 __all__ = [
@@ -25,6 +39,7 @@ __all__ = [
     "db",
     "transactional_db",
     "django_db_reset_sequences",
+    "django_db_serialized_rollback",
     "admin_user",
     "django_user_model",
     "django_username_field",
@@ -53,7 +68,7 @@ def django_db_modify_db_settings_tox_suffix() -> None:
 
 
 @pytest.fixture(scope="session")
-def django_db_modify_db_settings_xdist_suffix(request) -> None:
+def django_db_modify_db_settings_xdist_suffix(request: pytest.FixtureRequest) -> None:
     skip_if_no_django()
 
     xdist_suffix = getattr(request.config, "workerinput", {}).get("workerid")
@@ -78,37 +93,39 @@ def django_db_modify_db_settings(
 
 
 @pytest.fixture(scope="session")
-def django_db_use_migrations(request) -> bool:
+def django_db_use_migrations(request: pytest.FixtureRequest) -> bool:
     return not request.config.getvalue("nomigrations")
 
 
 @pytest.fixture(scope="session")
-def django_db_keepdb(request) -> bool:
-    return request.config.getvalue("reuse_db")
+def django_db_keepdb(request: pytest.FixtureRequest) -> bool:
+    reuse_db: bool = request.config.getvalue("reuse_db")
+    return reuse_db
 
 
 @pytest.fixture(scope="session")
-def django_db_createdb(request) -> bool:
-    return request.config.getvalue("create_db")
+def django_db_createdb(request: pytest.FixtureRequest) -> bool:
+    create_db: bool = request.config.getvalue("create_db")
+    return create_db
 
 
 @pytest.fixture(scope="session")
 def django_db_setup(
-    request,
+    request: pytest.FixtureRequest,
     django_test_environment: None,
     django_db_blocker,
     django_db_use_migrations: bool,
     django_db_keepdb: bool,
     django_db_createdb: bool,
     django_db_modify_db_settings: None,
-) -> None:
+) -> Generator[None, None, None]:
     """Top level fixture to ensure test databases are available"""
     from django.test.utils import setup_databases, teardown_databases
 
     setup_databases_args = {}
 
     if not django_db_use_migrations:
-        _disable_native_migrations()
+        _disable_migrations()
 
     if django_db_keepdb and not django_db_createdb:
         setup_databases_args["keepdb"] = True
@@ -117,46 +134,65 @@ def django_db_setup(
         db_cfg = setup_databases(
             verbosity=request.config.option.verbose,
             interactive=False,
-            **setup_databases_args
+            **setup_databases_args,
         )
 
-    def teardown_database() -> None:
+    yield
+
+    if not django_db_keepdb:
         with django_db_blocker.unblock():
             try:
                 teardown_databases(db_cfg, verbosity=request.config.option.verbose)
-            except Exception as exc:
+            except Exception as exc:  # noqa: BLE001
                 request.node.warn(
-                    pytest.PytestWarning(
-                        "Error when trying to teardown test databases: %r" % exc
-                    )
+                    pytest.PytestWarning(f"Error when trying to teardown test databases: {exc!r}")
                 )
 
-    if not django_db_keepdb:
-        request.addfinalizer(teardown_database)
 
-
-def _django_db_fixture_helper(
-    request,
+@pytest.fixture()
+def _django_db_helper(
+    request: pytest.FixtureRequest,
+    django_db_setup: None,
     django_db_blocker,
-    transactional: bool = False,
-    reset_sequences: bool = False,
-) -> None:
+) -> Generator[None, None, None]:
+    from django import VERSION
+
     if is_django_unittest(request):
+        yield
         return
 
-    if not transactional and "live_server" in request.fixturenames:
-        # Do nothing, we get called with transactional=True, too.
-        return
+    marker = request.node.get_closest_marker("django_db")
+    if marker:
+        (
+            transactional,
+            reset_sequences,
+            databases,
+            serialized_rollback,
+            available_apps,
+        ) = validate_django_db(marker)
+    else:
+        (
+            transactional,
+            reset_sequences,
+            databases,
+            serialized_rollback,
+            available_apps,
+        ) = False, False, None, False, None
 
-    _databases = getattr(
-        request.node, "_pytest_django_databases", None,
-    )  # type: Optional[_DjangoDbDatabases]
+    transactional = (
+        transactional
+        or reset_sequences
+        or ("transactional_db" in request.fixturenames or "live_server" in request.fixturenames)
+    )
+    reset_sequences = reset_sequences or ("django_db_reset_sequences" in request.fixturenames)
+    serialized_rollback = serialized_rollback or (
+        "django_db_serialized_rollback" in request.fixturenames
+    )
 
     django_db_blocker.unblock()
-    request.addfinalizer(django_db_blocker.restore)
 
-    import django.test
     import django.db
+    import django.test
 
     if transactional:
         test_case_class = django.test.TransactionTestCase
@@ -164,22 +200,87 @@ def _django_db_fixture_helper(
         test_case_class = django.test.TestCase
 
     _reset_sequences = reset_sequences
+    _serialized_rollback = serialized_rollback
+    _databases = databases
+    _available_apps = available_apps
 
     class PytestDjangoTestCase(test_case_class):  # type: ignore[misc,valid-type]
-        if transactional and _reset_sequences:
-            reset_sequences = True
+        reset_sequences = _reset_sequences
+        serialized_rollback = _serialized_rollback
         if _databases is not None:
             databases = _databases
+        if _available_apps is not None:
+            available_apps = _available_apps
+
+        # For non-transactional tests, skip executing `django.test.TestCase`'s
+        # `setUpClass`/`tearDownClass`, only execute the super class ones.
+        #
+        # `TestCase`'s class setup manages the `setUpTestData`/class-level
+        # transaction functionality. We don't use it; instead we (will) offer
+        # our own alternatives. So it only adds overhead, and does some things
+        # which conflict with our (planned) functionality, particularly, it
+        # closes all database connections in `tearDownClass` which inhibits
+        # wrapping tests in higher-scoped transactions.
+        #
+        # It's possible a new version of Django will add some unrelated
+        # functionality to these methods, in which case skipping them completely
+        # would not be desirable. Let's cross that bridge when we get there...
+        if not transactional:
+
+            @classmethod
+            def setUpClass(cls) -> None:
+                super(django.test.TestCase, cls).setUpClass()
+                if VERSION < (4, 1):
+                    django.db.transaction.Atomic._ensure_durability = False
+
+            @classmethod
+            def tearDownClass(cls) -> None:
+                if VERSION < (4, 1):
+                    django.db.transaction.Atomic._ensure_durability = True
+                super(django.test.TestCase, cls).tearDownClass()
 
     PytestDjangoTestCase.setUpClass()
-    request.addfinalizer(PytestDjangoTestCase.tearDownClass)
 
     test_case = PytestDjangoTestCase(methodName="__init__")
     test_case._pre_setup()
-    request.addfinalizer(test_case._post_teardown)
+
+    yield
+
+    test_case._post_teardown()
+
+    PytestDjangoTestCase.tearDownClass()
+
+    if VERSION >= (4, 0):
+        PytestDjangoTestCase.doClassCleanups()
+
+    django_db_blocker.restore()
 
 
-def _disable_native_migrations() -> None:
+def validate_django_db(marker) -> _DjangoDb:
+    """Validate the django_db marker.
+
+    It checks the signature and creates the ``transaction``,
+    ``reset_sequences``, ``databases``, ``serialized_rollback`` and
+    ``available_apps`` attributes on the marker which will have the correct
+    values.
+
+    Sequence reset, serialized_rollback, and available_apps are only allowed
+    when combined with transaction.
+    """
+
+    def apifun(
+        transaction: bool = False,
+        reset_sequences: bool = False,
+        databases: _DjangoDbDatabases = None,
+        serialized_rollback: bool = False,
+        available_apps: _DjangoDbAvailableApps = None,
+    ) -> _DjangoDb:
+        return transaction, reset_sequences, databases, serialized_rollback, available_apps
+
+    return apifun(*marker.args, **marker.kwargs)
+
+
+def _disable_migrations() -> None:
     from django.conf import settings
     from django.core.management.commands import migrate
 
@@ -209,54 +310,37 @@ def _set_suffix_to_test_databases(suffix: str) -> None:
         if not test_name:
             if db_settings["ENGINE"] == "django.db.backends.sqlite3":
                 continue
-            test_name = "test_{}".format(db_settings["NAME"])
+            test_name = f"test_{db_settings['NAME']}"
 
         if test_name == ":memory:":
             continue
 
         db_settings.setdefault("TEST", {})
-        db_settings["TEST"]["NAME"] = "{}_{}".format(test_name, suffix)
+        db_settings["TEST"]["NAME"] = f"{test_name}_{suffix}"
 
 
 # ############### User visible fixtures ################
 
 
-@pytest.fixture(scope="function")
-def db(
-    request,
-    django_db_setup: None,
-    django_db_blocker,
-) -> None:
+@pytest.fixture()
+def db(_django_db_helper: None) -> None:
     """Require a django test database.
 
     This database will be setup with the default fixtures and will have
     the transaction management disabled. At the end of the test the outer
     transaction that wraps the test itself will be rolled back to undo any
     changes to the database (in case the backend supports transactions).
-    This is more limited than the ``transactional_db`` resource but
+    This is more limited than the ``transactional_db`` fixture but
     faster.
 
-    If multiple database fixtures are requested, they take precedence
-    over each other in the following order (the last one wins): ``db``,
-    ``transactional_db``, ``django_db_reset_sequences``.
+    If both ``db`` and ``transactional_db`` are requested,
+    ``transactional_db`` takes precedence.
     """
-    if "django_db_reset_sequences" in request.fixturenames:
-        request.getfixturevalue("django_db_reset_sequences")
-    if (
-        "transactional_db" in request.fixturenames
-        or "live_server" in request.fixturenames
-    ):
-        request.getfixturevalue("transactional_db")
-    else:
-        _django_db_fixture_helper(request, django_db_blocker, transactional=False)
+    # The `_django_db_helper` fixture checks if `db` is requested.
 
 
-@pytest.fixture(scope="function")
-def transactional_db(
-    request,
-    django_db_setup: None,
-    django_db_blocker,
-) -> None:
+@pytest.fixture()
+def transactional_db(_django_db_helper: None) -> None:
     """Require a django test database with transaction support.
 
     This will re-initialise the django database for each test and is
@@ -265,39 +349,51 @@ def transactional_db(
     If you want to use the database with transactions you must request
     this resource.
 
-    If multiple database fixtures are requested, they take precedence
-    over each other in the following order (the last one wins): ``db``,
-    ``transactional_db``, ``django_db_reset_sequences``.
+    If both ``db`` and ``transactional_db`` are requested,
+    ``transactional_db`` takes precedence.
     """
-    if "django_db_reset_sequences" in request.fixturenames:
-        request.getfixturevalue("django_db_reset_sequences")
-    _django_db_fixture_helper(request, django_db_blocker, transactional=True)
-
-
-@pytest.fixture(scope="function")
-def django_db_reset_sequences(
-    request,
-    django_db_setup: None,
-    django_db_blocker,
-) -> None:
-    """Require a transactional test database with sequence reset support.
-
-    This behaves like the ``transactional_db`` fixture, with the addition
-    of enforcing a reset of all auto increment sequences.  If the enquiring
-    test relies on such values (e.g. ids as primary keys), you should
-    request this resource to ensure they are consistent across tests.
-
-    If multiple database fixtures are requested, they take precedence
-    over each other in the following order (the last one wins): ``db``,
-    ``transactional_db``, ``django_db_reset_sequences``.
-    """
-    _django_db_fixture_helper(
-        request, django_db_blocker, transactional=True, reset_sequences=True
-    )
+    # The `_django_db_helper` fixture checks if `transactional_db` is requested.
 
 
 @pytest.fixture()
-def client() -> "django.test.client.Client":
+def django_db_reset_sequences(
+    _django_db_helper: None,
+    transactional_db: None,
+) -> None:
+    """Require a transactional test database with sequence reset support.
+
+    This requests the ``transactional_db`` fixture, and additionally
+    enforces a reset of all auto increment sequences.  If the enquiring
+    test relies on such values (e.g. ids as primary keys), you should
+    request this resource to ensure they are consistent across tests.
+    """
+    # The `_django_db_helper` fixture checks if `django_db_reset_sequences`
+    # is requested.
+
+
+@pytest.fixture()
+def django_db_serialized_rollback(
+    _django_db_helper: None,
+    db: None,
+) -> None:
+    """Require a test database with serialized rollbacks.
+
+    This requests the ``db`` fixture, and additionally performs rollback
+    emulation - serializes the database contents during setup and restores
+    it during teardown.
+
+    This fixture may be useful for transactional tests, so is usually combined
+    with ``transactional_db``, but can also be useful on databases which do not
+    support transactions.
+
+    Note that this will slow down that test suite by approximately 3x.
+    """
+    # The `_django_db_helper` fixture checks if `django_db_serialized_rollback`
+    # is requested.
+
+
+@pytest.fixture()
+def client() -> django.test.client.Client:
     """A Django test client instance."""
     skip_if_no_django()
 
@@ -307,7 +403,7 @@ def client() -> "django.test.client.Client":
 
 
 @pytest.fixture()
-def async_client() -> "django.test.client.AsyncClient":
+def async_client() -> django.test.client.AsyncClient:
     """A Django test async client instance."""
     skip_if_no_django()
 
@@ -327,7 +423,8 @@ def django_user_model(db: None):
 @pytest.fixture()
 def django_username_field(django_user_model) -> str:
     """The fieldname for the username used with Django's user model."""
-    return django_user_model.USERNAME_FIELD
+    field: str = django_user_model.USERNAME_FIELD
+    return field
 
 
 @pytest.fixture()
@@ -365,7 +462,7 @@ def admin_user(
 def admin_client(
     db: None,
     admin_user,
-) -> "django.test.client.Client":
+) -> django.test.client.Client:
     """A Django test client logged in as an admin user."""
     from django.test.client import Client
 
@@ -375,7 +472,7 @@ def admin_client(
 
 
 @pytest.fixture()
-def rf() -> "django.test.client.RequestFactory":
+def rf() -> django.test.client.RequestFactory:
     """RequestFactory instance"""
     skip_if_no_django()
 
@@ -385,7 +482,7 @@ def rf() -> "django.test.client.RequestFactory":
 
 
 @pytest.fixture()
-def async_rf() -> "django.test.client.AsyncRequestFactory":
+def async_rf() -> django.test.client.AsyncRequestFactory:
     """AsyncRequestFactory instance"""
     skip_if_no_django()
 
@@ -395,7 +492,7 @@ def async_rf() -> "django.test.client.AsyncRequestFactory":
 
 
 class SettingsWrapper:
-    _to_restore = []  # type: List[Any]
+    _to_restore: ClassVar[list[Any]] = []
 
     def __delattr__(self, attr: str) -> None:
         from django.test import override_settings
@@ -438,7 +535,7 @@ def settings():
 
 
 @pytest.fixture(scope="session")
-def live_server(request):
+def live_server(request: pytest.FixtureRequest):
     """Run a live Django server in the background during tests
 
     The address the server is started from is taken from the
@@ -459,17 +556,19 @@ def live_server(request):
     """
     skip_if_no_django()
 
-    addr = request.config.getvalue("liveserver") or os.getenv(
-        "DJANGO_LIVE_TEST_SERVER_ADDRESS"
-    ) or "localhost"
+    addr = (
+        request.config.getvalue("liveserver")
+        or os.getenv("DJANGO_LIVE_TEST_SERVER_ADDRESS")
+        or "localhost"
+    )
 
     server = live_server_helper.LiveServer(addr)
-    request.addfinalizer(server.stop)
-    return server
+    yield server
+    server.stop()
 
 
-@pytest.fixture(autouse=True, scope="function")
-def _live_server_helper(request) -> None:
+@pytest.fixture(autouse=True)
+def _live_server_helper(request: pytest.FixtureRequest) -> Generator[None, None, None]:
     """Helper to make live_server work, internal to pytest-django.
 
     This helper will dynamically request the transactional_db fixture
@@ -485,13 +584,15 @@ def _live_server_helper(request) -> None:
     It will also override settings only for the duration of the test.
     """
     if "live_server" not in request.fixturenames:
+        yield
         return
 
     request.getfixturevalue("transactional_db")
 
     live_server = request.getfixturevalue("live_server")
     live_server._live_server_modified_settings.enable()
-    request.addfinalizer(live_server._live_server_modified_settings.disable)
+    yield
+    live_server._live_server_modified_settings.disable()
 
 
 @contextmanager
@@ -501,7 +602,7 @@ def _assert_num_queries(
     exact: bool = True,
     connection=None,
     info=None,
-) -> Generator["django.test.utils.CaptureQueriesContext", None, None]:
+) -> Generator[django.test.utils.CaptureQueriesContext, None, None]:
     from django.test.utils import CaptureQueriesContext
 
     if connection is None:
@@ -518,15 +619,13 @@ def _assert_num_queries(
         else:
             failed = num_performed > num
         if failed:
-            msg = "Expected to perform {} queries {}{}".format(
-                num,
-                "" if exact else "or less ",
-                "but {} done".format(
-                    num_performed == 1 and "1 was" or "{} were".format(num_performed)
-                ),
-            )
+            msg = f"Expected to perform {num} queries "
+            if not exact:
+                msg += "or less "
+            verb = "was" if num_performed == 1 else "were"
+            msg += f"but {num_performed} {verb} done"
             if info:
-                msg += "\n{}".format(info)
+                msg += f"\n{info}"
             if verbose:
                 sqls = (q["sql"] for q in context.captured_queries)
                 msg += "\n\nQueries:\n========\n\n" + "\n\n".join(sqls)
@@ -535,46 +634,18 @@ def _assert_num_queries(
             pytest.fail(msg)
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture()
 def django_assert_num_queries(pytestconfig):
     return partial(_assert_num_queries, pytestconfig)
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture()
 def django_assert_max_num_queries(pytestconfig):
     return partial(_assert_num_queries, pytestconfig, exact=False)
 
 
-@contextmanager
-def _capture_on_commit_callbacks(
-    *,
-    using: Optional[str] = None,
-    execute: bool = False
-):
-    from django.db import DEFAULT_DB_ALIAS, connections
+@pytest.fixture()
+def django_capture_on_commit_callbacks():
     from django.test import TestCase
 
-    if using is None:
-        using = DEFAULT_DB_ALIAS
-
-    # Polyfill of Django code as of Django 3.2.
-    if get_django_version() < (3, 2):
-        callbacks = []  # type: List[Callable[[], Any]]
-        start_count = len(connections[using].run_on_commit)
-        try:
-            yield callbacks
-        finally:
-            run_on_commit = connections[using].run_on_commit[start_count:]
-            callbacks[:] = [func for sids, func in run_on_commit]
-            if execute:
-                for callback in callbacks:
-                    callback()
-
-    else:
-        with TestCase.captureOnCommitCallbacks(using=using, execute=execute) as callbacks:
-            yield callbacks
-
-
-@pytest.fixture(scope="function")
-def django_capture_on_commit_callbacks():
-    return _capture_on_commit_callbacks
+    return TestCase.captureOnCommitCallbacks
