@@ -3,6 +3,7 @@
 This plugin handles creating and destroying the test environment and
 test database and provides some useful text fixtures.
 """
+from __future__ import annotations
 
 import contextlib
 import inspect
@@ -10,46 +11,45 @@ import os
 import pathlib
 import sys
 from functools import reduce
-from typing import Generator, List, Optional, Tuple, Union
+from typing import TYPE_CHECKING, ContextManager, Generator, List, NoReturn
 
 import pytest
 
-from .django_compat import is_django_unittest  # noqa
-from .fixtures import _django_db_helper  # noqa
-from .fixtures import _live_server_helper  # noqa
-from .fixtures import admin_client  # noqa
-from .fixtures import admin_user  # noqa
-from .fixtures import async_client  # noqa
-from .fixtures import async_rf  # noqa
-from .fixtures import client  # noqa
-from .fixtures import db  # noqa
-from .fixtures import django_assert_max_num_queries  # noqa
-from .fixtures import django_assert_num_queries  # noqa
-from .fixtures import django_capture_on_commit_callbacks  # noqa
-from .fixtures import django_db_createdb  # noqa
-from .fixtures import django_db_keepdb  # noqa
-from .fixtures import django_db_modify_db_settings  # noqa
-from .fixtures import django_db_modify_db_settings_parallel_suffix  # noqa
-from .fixtures import django_db_modify_db_settings_tox_suffix  # noqa
-from .fixtures import django_db_modify_db_settings_xdist_suffix  # noqa
-from .fixtures import django_db_reset_sequences  # noqa
-from .fixtures import django_db_serialized_rollback  # noqa
-from .fixtures import django_db_setup  # noqa
-from .fixtures import django_db_use_migrations  # noqa
-from .fixtures import django_user_model  # noqa
-from .fixtures import django_username_field  # noqa
-from .fixtures import live_server  # noqa
-from .fixtures import rf  # noqa
-from .fixtures import settings  # noqa
-from .fixtures import transactional_db  # noqa
-from .fixtures import validate_django_db
+from .django_compat import is_django_unittest
+from .fixtures import (
+    _django_db_helper,  # noqa: F401
+    _live_server_helper,  # noqa: F401
+    admin_client,  # noqa: F401
+    admin_user,  # noqa: F401
+    async_client,  # noqa: F401
+    async_rf,  # noqa: F401
+    client,  # noqa: F401
+    db,  # noqa: F401
+    django_assert_max_num_queries,  # noqa: F401
+    django_assert_num_queries,  # noqa: F401
+    django_capture_on_commit_callbacks,  # noqa: F401
+    django_db_createdb,  # noqa: F401
+    django_db_keepdb,  # noqa: F401
+    django_db_modify_db_settings,  # noqa: F401
+    django_db_modify_db_settings_parallel_suffix,  # noqa: F401
+    django_db_modify_db_settings_tox_suffix,  # noqa: F401
+    django_db_modify_db_settings_xdist_suffix,  # noqa: F401
+    django_db_reset_sequences,  # noqa: F401
+    django_db_serialized_rollback,  # noqa: F401
+    django_db_setup,  # noqa: F401
+    django_db_use_migrations,  # noqa: F401
+    django_user_model,  # noqa: F401
+    django_username_field,  # noqa: F401
+    live_server,  # noqa: F401
+    rf,  # noqa: F401
+    settings,  # noqa: F401
+    transactional_db,  # noqa: F401
+    validate_django_db,
+)
 from .lazy_django import django_settings_is_configured, skip_if_no_django
 
 
-TYPE_CHECKING = False
 if TYPE_CHECKING:
-    from typing import ContextManager, NoReturn
-
     import django
 
 
@@ -57,14 +57,12 @@ SETTINGS_MODULE_ENV = "DJANGO_SETTINGS_MODULE"
 CONFIGURATION_ENV = "DJANGO_CONFIGURATION"
 INVALID_TEMPLATE_VARS_ENV = "FAIL_INVALID_TEMPLATE_VARS"
 
-_report_header = []
-
 
 # ############### pytest hooks ################
 
 
 @pytest.hookimpl()
-def pytest_addoption(parser) -> None:
+def pytest_addoption(parser: pytest.Parser) -> None:
     group = parser.getgroup("django")
     group.addoption(
         "--reuse-db",
@@ -114,7 +112,8 @@ def pytest_addoption(parser) -> None:
         help="Enable Django migrations on test setup",
     )
     parser.addini(
-        CONFIGURATION_ENV, "django-configurations class to use by pytest-django."
+        CONFIGURATION_ENV,
+        "django-configurations class to use by pytest-django.",
     )
     group.addoption(
         "--liveserver",
@@ -122,7 +121,8 @@ def pytest_addoption(parser) -> None:
         help="Address and port for the live_server fixture.",
     )
     parser.addini(
-        SETTINGS_MODULE_ENV, "Django settings module to use by pytest-django."
+        SETTINGS_MODULE_ENV,
+        "Django settings module to use by pytest-django.",
     )
 
     parser.addini(
@@ -133,8 +133,7 @@ def pytest_addoption(parser) -> None:
     )
     parser.addini(
         "django_debug_mode",
-        "How to set the Django DEBUG setting (default `False`). "
-        "Use `keep` to not override.",
+        "How to set the Django DEBUG setting (default `False`). " "Use `keep` to not override.",
         default="False",
     )
     group.addoption(
@@ -180,7 +179,7 @@ def _handle_import_error(extra_message: str) -> Generator[None, None, None]:
     except ImportError as e:
         django_msg = (e.args[0] + "\n\n") if e.args else ""
         msg = django_msg + extra_message
-        raise ImportError(msg)
+        raise ImportError(msg) from None
 
 
 def _add_django_project_to_path(args) -> str:
@@ -195,7 +194,7 @@ def _add_django_project_to_path(args) -> str:
         arg = arg.split("::", 1)[0]
         return pathlib.Path(arg)
 
-    def find_django_path(args) -> Optional[pathlib.Path]:
+    def find_django_path(args) -> pathlib.Path | None:
         str_args = (str(arg) for arg in args)
         path_args = [arg_to_path(x) for x in str_args if not x.startswith("-")]
 
@@ -239,9 +238,9 @@ def _setup_django() -> None:
 
 
 def _get_boolean_value(
-    x: Union[None, bool, str],
+    x: None | (bool | str),
     name: str,
-    default: Optional[bool] = None,
+    default: bool | None = None,
 ) -> bool:
     if x is None:
         return bool(default)
@@ -251,17 +250,20 @@ def _get_boolean_value(
     try:
         return possible_values[x.lower()]
     except KeyError:
+        possible = ", ".join(possible_values)
         raise ValueError(
-            "{} is not a valid value for {}. "
-            "It must be one of {}.".format(x, name, ", ".join(possible_values.keys()))
-        )
+            f"{x} is not a valid value for {name}. It must be one of {possible}."
+        ) from None
+
+
+report_header_key = pytest.StashKey[List[str]]()
 
 
 @pytest.hookimpl()
 def pytest_load_initial_conftests(
-    early_config,
-    parser,
-    args: List[str],
+    early_config: pytest.Config,
+    parser: pytest.Parser,
+    args: list[str],
 ) -> None:
     # Register the marks
     early_config.addinivalue_line(
@@ -307,17 +309,15 @@ def pytest_load_initial_conftests(
 
     if (
         options.itv
-        or _get_boolean_value(
-            os.environ.get(INVALID_TEMPLATE_VARS_ENV), INVALID_TEMPLATE_VARS_ENV
-        )
+        or _get_boolean_value(os.environ.get(INVALID_TEMPLATE_VARS_ENV), INVALID_TEMPLATE_VARS_ENV)
         or early_config.getini(INVALID_TEMPLATE_VARS_ENV)
     ):
         os.environ[INVALID_TEMPLATE_VARS_ENV] = "true"
 
     def _get_option_with_source(
-        option: Optional[str],
+        option: str | None,
         envname: str,
-    ) -> Union[Tuple[str, str], Tuple[None, None]]:
+    ) -> tuple[str, str] | tuple[None, None]:
         if option:
             return option, "option"
         if envname in os.environ:
@@ -330,12 +330,15 @@ def pytest_load_initial_conftests(
     ds, ds_source = _get_option_with_source(options.ds, SETTINGS_MODULE_ENV)
     dc, dc_source = _get_option_with_source(options.dc, CONFIGURATION_ENV)
 
+    report_header: list[str] = []
+    early_config.stash[report_header_key] = report_header
+
     if ds:
-        _report_header.append("settings: {} (from {})".format(ds, ds_source))
+        report_header.append(f"settings: {ds} (from {ds_source})")
         os.environ[SETTINGS_MODULE_ENV] = ds
 
         if dc:
-            _report_header.append("configuration: {} (from {})".format(dc, dc_source))
+            report_header.append(f"configuration: {dc} (from {dc_source})")
             os.environ[CONFIGURATION_ENV] = dc
 
             # Install the django-configurations importer
@@ -348,18 +351,11 @@ def pytest_load_initial_conftests(
         from django.conf import settings as dj_settings
 
         with _handle_import_error(_django_project_scan_outcome):
-            dj_settings.DATABASES
+            dj_settings.DATABASES  # noqa: B018
 
     _setup_django()
 
 
-@pytest.hookimpl()
-def pytest_report_header() -> Optional[List[str]]:
-    if _report_header:
-        return ["django: " + ", ".join(_report_header)]
-    return None
-
-
 @pytest.hookimpl(trylast=True)
 def pytest_configure() -> None:
     # Allow Django settings to be configured in a user pytest_configure call,
@@ -367,8 +363,64 @@ def pytest_configure() -> None:
     _setup_django()
 
 
+@pytest.hookimpl()
+def pytest_report_header(config: pytest.Config) -> list[str] | None:
+    report_header = config.stash[report_header_key]
+
+    if "django" in sys.modules:
+        import django
+
+        report_header.insert(0, f"version: {django.get_version()}")
+
+    if report_header:
+        return ["django: " + ", ".join(report_header)]
+    return None
+
+
+# Convert Django test tags on test classes to pytest marks.
+def pytest_collectstart(collector: pytest.Collector) -> None:
+    if "django" not in sys.modules:
+        return
+
+    if not isinstance(collector, pytest.Class):
+        return
+
+    tags = getattr(collector.obj, "tags", ())
+    if not tags:
+        return
+
+    from django.test import TransactionTestCase
+
+    if not issubclass(collector.obj, TransactionTestCase):
+        return
+
+    for tag in tags:
+        collector.add_marker(tag)
+
+
+# Convert Django test tags on test methods to pytest marks.
+def pytest_itemcollected(item: pytest.Item) -> None:
+    if "django" not in sys.modules:
+        return
+
+    if not isinstance(item, pytest.Function):
+        return
+
+    tags = getattr(item.obj, "tags", ())
+    if not tags:
+        return
+
+    from django.test import TransactionTestCase
+
+    if not issubclass(item.cls, TransactionTestCase):
+        return
+
+    for tag in tags:
+        item.add_marker(tag)
+
+
 @pytest.hookimpl(tryfirst=True)
-def pytest_collection_modifyitems(items: List[pytest.Item]) -> None:
+def pytest_collection_modifyitems(items: list[pytest.Item]) -> None:
     # If Django is not configured we don't need to bother
     if not django_settings_is_configured():
         return
@@ -389,6 +441,7 @@ def pytest_collection_modifyitems(items: List[pytest.Item]) -> None:
                     reset_sequences,
                     databases,
                     serialized_rollback,
+                    available_apps,
                 ) = validate_django_db(marker_db)
                 uses_db = True
                 transactional = transaction or reset_sequences
@@ -410,7 +463,7 @@ def pytest_collection_modifyitems(items: List[pytest.Item]) -> None:
 
 
 @pytest.fixture(autouse=True, scope="session")
-def django_test_environment(request) -> None:
+def django_test_environment(request: pytest.FixtureRequest) -> Generator[None, None, None]:
     """
     Ensure that Django is loaded and has its testing environment setup.
 
@@ -423,9 +476,7 @@ def django_test_environment(request) -> None:
     """
     if django_settings_is_configured():
         _setup_django()
-        from django.test.utils import (
-            setup_test_environment, teardown_test_environment,
-        )
+        from django.test.utils import setup_test_environment, teardown_test_environment
 
         debug_ini = request.config.getini("django_debug_mode")
         if debug_ini == "keep":
@@ -434,11 +485,15 @@ def django_test_environment(request) -> None:
             debug = _get_boolean_value(debug_ini, "django_debug_mode", False)
 
         setup_test_environment(debug=debug)
-        request.addfinalizer(teardown_test_environment)
+        yield
+        teardown_test_environment()
+
+    else:
+        yield
 
 
 @pytest.fixture(scope="session")
-def django_db_blocker() -> "Optional[_DatabaseBlocker]":
+def django_db_blocker() -> _DatabaseBlocker | None:
     """Wrapper around Django's database access.
 
     This object can be used to re-enable database access.  This fixture is used
@@ -458,7 +513,7 @@ def django_db_blocker() -> "Optional[_DatabaseBlocker]":
 
 
 @pytest.fixture(autouse=True)
-def _django_db_marker(request) -> None:
+def _django_db_marker(request: pytest.FixtureRequest) -> None:
     """Implement the django_db marker, internal to pytest-django."""
     marker = request.node.get_closest_marker("django_db")
     if marker:
@@ -467,8 +522,8 @@ def _django_db_marker(request) -> None:
 
 @pytest.fixture(autouse=True, scope="class")
 def _django_setup_unittest(
-    request,
-    django_db_blocker: "_DatabaseBlocker",
+    request: pytest.FixtureRequest,
+    django_db_blocker: _DatabaseBlocker,
 ) -> Generator[None, None, None]:
     """Setup a django unittest, internal to pytest-django."""
     if not django_settings_is_configured() or not is_django_unittest(request):
@@ -479,23 +534,24 @@ def _django_setup_unittest(
     # Before pytest 5.4: https://github.com/pytest-dev/pytest/issues/5991
     # After pytest 5.4: https://github.com/pytest-dev/pytest-django/issues/824
     from _pytest.unittest import TestCaseFunction
+
     original_runtest = TestCaseFunction.runtest
 
     def non_debugging_runtest(self) -> None:
         self._testcase(result=self)
 
     try:
-        TestCaseFunction.runtest = non_debugging_runtest  # type: ignore[assignment]
+        TestCaseFunction.runtest = non_debugging_runtest  # type: ignore[method-assign]
 
         request.getfixturevalue("django_db_setup")
 
         with django_db_blocker.unblock():
             yield
     finally:
-        TestCaseFunction.runtest = original_runtest  # type: ignore[assignment]
+        TestCaseFunction.runtest = original_runtest  # type: ignore[method-assign]
 
 
-@pytest.fixture(scope="function", autouse=True)
+@pytest.fixture(autouse=True)
 def _dj_autoclear_mailbox() -> None:
     if not django_settings_is_configured():
         return
@@ -505,22 +561,22 @@ def _dj_autoclear_mailbox() -> None:
     del mail.outbox[:]
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture()
 def mailoutbox(
     django_mail_patch_dns: None,
     _dj_autoclear_mailbox: None,
-) -> "Optional[List[django.core.mail.EmailMessage]]":
+) -> list[django.core.mail.EmailMessage] | None:
     if not django_settings_is_configured():
         return None
 
     from django.core import mail
 
-    return mail.outbox
+    return mail.outbox  # type: ignore[no-any-return]
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture()
 def django_mail_patch_dns(
-    monkeypatch,
+    monkeypatch: pytest.MonkeyPatch,
     django_mail_dnsname: str,
 ) -> None:
     from django.core import mail
@@ -528,13 +584,13 @@ def django_mail_patch_dns(
     monkeypatch.setattr(mail.message, "DNS_NAME", django_mail_dnsname)
 
 
-@pytest.fixture(scope="function")
+@pytest.fixture()
 def django_mail_dnsname() -> str:
     return "fake-tests.example.com"
 
 
-@pytest.fixture(autouse=True, scope="function")
-def _django_set_urlconf(request) -> None:
+@pytest.fixture(autouse=True)
+def _django_set_urlconf(request: pytest.FixtureRequest) -> Generator[None, None, None]:
     """Apply the @pytest.mark.urls marker, internal to pytest-django."""
     marker = request.node.get_closest_marker("urls")
     if marker:
@@ -548,14 +604,14 @@ def _django_set_urlconf(request) -> None:
         clear_url_caches()
         set_urlconf(None)
 
-        def restore() -> None:
-            django.conf.settings.ROOT_URLCONF = original_urlconf
-            # Copy the pattern from
-            # https://github.com/django/django/blob/main/django/test/signals.py#L152
-            clear_url_caches()
-            set_urlconf(None)
+    yield
 
-        request.addfinalizer(restore)
+    if marker:
+        django.conf.settings.ROOT_URLCONF = original_urlconf
+        # Copy the pattern from
+        # https://github.com/django/django/blob/main/django/test/signals.py#L152
+        clear_url_caches()
+        set_urlconf(None)
 
 
 @pytest.fixture(autouse=True, scope="session")
@@ -583,7 +639,7 @@ def _fail_for_invalid_template_variable():
             return key == "%s"
 
         @staticmethod
-        def _get_origin():
+        def _get_origin() -> str | None:
             stack = inspect.stack()
 
             # Try to use topmost `self.origin` first (Django 1.9+, and with
@@ -592,10 +648,11 @@ def _fail_for_invalid_template_variable():
                 func = f[3]
                 if func == "render":
                     frame = f[0]
+                    origin: str | None
                     try:
                         origin = frame.f_locals["self"].origin
                     except (AttributeError, KeyError):
-                        continue
+                        origin = None
                     if origin is not None:
                         return origin
 
@@ -614,43 +671,58 @@ def _fail_for_invalid_template_variable():
             # ``django.template.base.Template``
             template = f_locals["self"]
             if isinstance(template, Template):
-                return template.name
+                name: str = template.name
+                return name
+            return None
 
         def __mod__(self, var: str) -> str:
             origin = self._get_origin()
             if origin:
-                msg = "Undefined template variable '{}' in '{}'".format(var, origin)
+                msg = f"Undefined template variable '{var}' in '{origin}'"
             else:
-                msg = "Undefined template variable '%s'" % var
+                msg = f"Undefined template variable '{var}'"
             if self.fail:
                 pytest.fail(msg)
             else:
                 return msg
 
-    if (
-        os.environ.get(INVALID_TEMPLATE_VARS_ENV, "false") == "true"
-        and django_settings_is_configured()
-    ):
-        from django.conf import settings as dj_settings
+    with pytest.MonkeyPatch.context() as mp:
+        if (
+            os.environ.get(INVALID_TEMPLATE_VARS_ENV, "false") == "true"
+            and django_settings_is_configured()
+        ):
+            from django.conf import settings as dj_settings
 
-        if dj_settings.TEMPLATES:
-            dj_settings.TEMPLATES[0]["OPTIONS"]["string_if_invalid"] = InvalidVarException()
+            if dj_settings.TEMPLATES:
+                mp.setitem(
+                    dj_settings.TEMPLATES[0]["OPTIONS"],
+                    "string_if_invalid",
+                    InvalidVarException(),
+                )
+        yield
 
 
 @pytest.fixture(autouse=True)
-def _template_string_if_invalid_marker(request) -> None:
+def _template_string_if_invalid_marker(
+    monkeypatch: pytest.MonkeyPatch,
+    request: pytest.FixtureRequest,
+) -> None:
     """Apply the @pytest.mark.ignore_template_errors marker,
-     internal to pytest-django."""
+    internal to pytest-django."""
     marker = request.keywords.get("ignore_template_errors", None)
     if os.environ.get(INVALID_TEMPLATE_VARS_ENV, "false") == "true":
         if marker and django_settings_is_configured():
             from django.conf import settings as dj_settings
 
             if dj_settings.TEMPLATES:
-                dj_settings.TEMPLATES[0]["OPTIONS"]["string_if_invalid"].fail = False
+                monkeypatch.setattr(
+                    dj_settings.TEMPLATES[0]["OPTIONS"]["string_if_invalid"],
+                    "fail",
+                    False,
+                )
 
 
-@pytest.fixture(autouse=True, scope="function")
+@pytest.fixture(autouse=True)
 def _django_clear_site_cache() -> None:
     """Clears ``django.contrib.sites.models.SITE_CACHE`` to avoid
     unexpected behavior with cached site objects.
@@ -685,12 +757,12 @@ class _DatabaseBlocker:
     This is the object returned by django_db_blocker.
     """
 
-    def __init__(self):
-        self._history = []
+    def __init__(self) -> None:
+        self._history = []  # type: ignore[var-annotated]
         self._real_ensure_connection = None
 
     @property
-    def _dj_db_wrapper(self) -> "django.db.backends.base.base.BaseDatabaseWrapper":
+    def _dj_db_wrapper(self) -> django.db.backends.base.base.BaseDatabaseWrapper:
         from django.db.backends.base.base import BaseDatabaseWrapper
 
         # The first time the _dj_db_wrapper is accessed, we will save a
@@ -703,22 +775,21 @@ class _DatabaseBlocker:
     def _save_active_wrapper(self) -> None:
         self._history.append(self._dj_db_wrapper.ensure_connection)
 
-    def _blocking_wrapper(*args, **kwargs) -> "NoReturn":
+    def _blocking_wrapper(*args, **kwargs) -> NoReturn:
         __tracebackhide__ = True
-        __tracebackhide__  # Silence pyflakes
         raise RuntimeError(
             "Database access not allowed, "
             'use the "django_db" mark, or the '
             '"db" or "transactional_db" fixtures to enable it.'
         )
 
-    def unblock(self) -> "ContextManager[None]":
+    def unblock(self) -> ContextManager[None]:
         """Enable access to the Django database."""
         self._save_active_wrapper()
         self._dj_db_wrapper.ensure_connection = self._real_ensure_connection
         return _DatabaseBlockerContextManager(self)
 
-    def block(self) -> "ContextManager[None]":
+    def block(self) -> ContextManager[None]:
         """Disable access to the Django database."""
         self._save_active_wrapper()
         self._dj_db_wrapper.ensure_connection = self._blocking_wrapper
@@ -731,14 +802,14 @@ class _DatabaseBlocker:
 _blocking_manager = _DatabaseBlocker()
 
 
-def validate_urls(marker) -> List[str]:
+def validate_urls(marker) -> list[str]:
     """Validate the urls marker.
 
     It checks the signature and creates the `urls` attribute on the
     marker which will have the correct value.
     """
 
-    def apifun(urls: List[str]) -> List[str]:
+    def apifun(urls: list[str]) -> list[str]:
         return urls
 
     return apifun(*marker.args, **marker.kwargs)
